from bluesky_live.run_builder import build_simple_run
import pytest

from ..plot_builders import Lines
from ..plot_specs import Axes, Figure


# Make some runs to use.
runs = [
    build_simple_run(
        {"motor": [1, 2], "det": [10, 20], "det2": [15, 25]},
        metadata={"scan_id": 1 + i},
    )
    for i in range(10)
]
MAX_RUNS = 3


def test_pinned(FigureView):
    "Test Lines with 'pinned' and un-pinned runs."
    ys = ["det", "det2"]
    num_ys = len(ys)
    model = Lines("motor", ys, max_runs=MAX_RUNS)
    view = FigureView(model.figure)

    # Add MAX_RUNS and then some more and check that they do get bumped off.
    for run in runs[:5]:
        model.add_run(run)
        assert len(model.runs) <= MAX_RUNS
    assert runs[2:5] == list(model.runs)

    # Add a pinned run.
    pinned_run = runs[5]
    model.add_run(pinned_run, pinned=True)
    assert frozenset([pinned_run.metadata["start"]["uid"]]) == model.pinned
    for run in runs[6:]:
        model.add_run(run)
        assert len(model.runs) == 1 + MAX_RUNS
        assert len(model.figure.axes[0].artists) == num_ys * (1 + MAX_RUNS)
    # Check that it hasn't been bumped off.
    assert pinned_run in model.runs

    # Remove the pinned run.
    model.discard_run(pinned_run)
    assert len(model.runs) == MAX_RUNS
    assert len(model.figure.axes[0].artists) == num_ys * MAX_RUNS
    assert pinned_run not in model.runs

    view.close()


def test_properties(FigureView):
    "Touch various accessors"
    model = Lines("c * motor", ["det"], namespace={"c": 3}, max_runs=MAX_RUNS)
    view = FigureView(model.figure)
    model.add_run(runs[0])
    assert model.runs[0] is runs[0]
    assert model.max_runs == MAX_RUNS
    assert model.x == "c * motor"
    assert list(model.ys) == ["det"]
    assert dict(model.namespace) == {"c": 3}
    assert model.needs_streams == ("primary",)
    assert model.pinned == frozenset()

    view.close()


@pytest.mark.parametrize(
    "operation,operation_args,num_ys,ys_list,num_lines",
    [
        ("append", ("det+1",), 2, ["det", "det+1"], 2),
        ("extend", (["det+1"],), 2, ["det", "det+1"], 2),
        ("insert", (0, "det+1"), 2, ["det+1", "det"], 2),
        ("insert", (1, "det+1"), 2, ["det", "det+1"], 2),
    ],
)
def test_adding_ys(operation, operation_args, num_ys, ys_list, num_lines, FigureView):
    "Test that append, extend, and insert work properly"
    model = Lines("c * motor", ["det"], namespace={"c": 3}, max_runs=MAX_RUNS)
    view = FigureView(model.figure)
    model.add_run(runs[0])
    assert len(model.ys) == 1
    assert list(model.ys) == ["det"]
    assert len(model.figure.axes[0].artists) == 1
    getattr(model.ys, operation)(*operation_args)
    assert len(model.ys) == num_ys
    assert list(model.ys) == ys_list
    assert len(model.figure.axes[0].artists) == num_lines
    view.close()


@pytest.mark.parametrize(
    "operation,operation_args,num_ys,ys_list,num_lines",
    [
        ("remove", ("det+1",), 2, ["det", "det+2"], 2),
        ("pop", (), 2, ["det", "det+1"], 2),
        ("pop", (1,), 2, ["det", "det+2"], 2),
        ("clear", (), 0, [], 0),
    ],
)
def test_removing_ys(operation, operation_args, num_ys, ys_list, num_lines, FigureView):
    "Test that remove, pop, del, and clear work properly"
    model = Lines("c * motor", ["det", "det+1", "det+2"], namespace={"c": 3}, max_runs=MAX_RUNS)
    view = FigureView(model.figure)
    model.add_run(runs[0])
    assert len(model.ys) == 3
    assert list(model.ys) == ["det", "det+1", "det+2"]
    assert len(model.figure.axes[0].artists) == 3
    getattr(model.ys, operation)(*operation_args)
    assert len(model.ys) == num_ys
    assert list(model.ys) == ys_list
    assert len(model.figure.axes[0].artists) == num_lines
    view.close()


def test_decrease_max_runs(FigureView):
    "Decreasing max_runs should remove the runs and their associated lines."
    INITIAL_MAX_RUNS = 5
    model = Lines("motor", ["det"], namespace={"c": 3}, max_runs=INITIAL_MAX_RUNS)
    view = FigureView(model.figure)
    for run in runs[:5]:
        model.add_run(run)
    assert len(model.runs) == INITIAL_MAX_RUNS
    assert len(model.figure.axes[0].artists) == INITIAL_MAX_RUNS
    # Decrease max_runs.
    model.max_runs = MAX_RUNS
    assert len(model.runs) == MAX_RUNS
    assert len(model.figure.axes[0].artists) == MAX_RUNS

    view.close()


@pytest.mark.parametrize("expr", ["det / det2", "-log(det)", "np.sqrt(det)"])
def test_expressions(expr, FigureView):
    "Test Lines with 'pinned' and un-pinned runs."
    ys = [expr]
    model = Lines("motor", ys, max_runs=MAX_RUNS)
    view = FigureView(model.figure)
    model.add_run(runs[0])
    assert len(model.figure.axes[0].artists) == 1

    view.close()


@pytest.mark.parametrize(
    "func",
    [
        lambda det, det2: det / det2,
        lambda det, log: -log(det),
        lambda det, np: np.sqrt(det),
    ],
    ids=["division", "top-level-log", "np-dot-log"],
)
def test_functions(func, FigureView):
    "Test Lines with 'pinned' and un-pinned runs."
    ys = [func]
    model = Lines("motor", ys, max_runs=MAX_RUNS)
    view = FigureView(model.figure)
    model.add_run(runs[0])
    assert len(model.figure.axes[0].artists) == 1

    view.close()


def test_figure_set_after_instantiation(FigureView):
    axes = Axes()
    model = Lines("motor", [], axes=axes)
    assert model.figure is None
    figure = Figure((axes,), title="")
    assert model.figure is figure
    view = FigureView(model.figure)
    view.close()


@pytest.mark.parametrize(
    "test_x_label,expected_x_label",
    [
        (None, "motor"),
        ("test", "test"),
    ],
)
def test_x_label(test_x_label, expected_x_label, FigureView):
    "Test that Lines properly sets the x_label."
    axes = Axes(x_label=test_x_label)
    model = Lines("motor", ["det"], axes=axes)
    figure = Figure((axes,), title="")
    view = FigureView(model.figure)
    assert model.axes.x_label == expected_x_label
    assert figure.axes[0].x_label == expected_x_label
    assert view.figure.axes[0].get_xlabel() == expected_x_label
    view.close()


@pytest.mark.parametrize(
    "test_y_labels,expected_y_labels",
    [
        ([None, "test"], ["det", "det, det+1", "test", "test"]),
        (["test", ""], ["test", "test", "", ""]),
        (["", None], ["", "", "det, det+1", "det, det+1, det+2"]),
    ],
)
def test_y_label(test_y_labels, expected_y_labels, FigureView):
    "Test that Lines correctly sets and updates the y_label."
    axes = Axes(y_label=test_y_labels[0])
    model = Lines("motor", ["det"], axes=axes)
    figure = Figure((axes,), title="")
    view = FigureView(model.figure)
    assert model.y_label == model.axes.y_label == expected_y_labels[0]
    assert figure.axes[0].y_label == expected_y_labels[0]
    assert view.figure.axes[0].get_ylabel() == expected_y_labels[0]
    model.ys.append("det+1")
    assert model.y_label == model.axes.y_label == expected_y_labels[1]
    assert figure.axes[0].y_label == expected_y_labels[1]
    assert view.figure.axes[0].get_ylabel() == expected_y_labels[1]
    model.y_label = test_y_labels[1]
    assert model.y_label == model.axes.y_label == expected_y_labels[2]
    assert figure.axes[0].y_label == expected_y_labels[2]
    assert view.figure.axes[0].get_ylabel() == expected_y_labels[2]
    model.ys.append("det+2")
    assert model.y_label == model.axes.y_label == expected_y_labels[3]
    assert figure.axes[0].y_label == expected_y_labels[3]
    assert view.figure.axes[0].get_ylabel() == expected_y_labels[3]

    view.close()


@pytest.mark.parametrize(
    "test_titles,expected_titles",
    [
        ([None, "test"], ["det v motor", "det, det+1 v motor", "test", "test"]),
        (["test", ""], ["test", "test", "", ""]),
        (["", None], ["", "", "det, det+1 v motor", "det, det+1, det+2 v motor"]),
    ],
)
def test_title(test_titles, expected_titles, FigureView):
    "Test that Lines correctly sets and updates the y_label."
    axes = Axes(title=test_titles[0])
    model = Lines("motor", ["det"], axes=axes)
    figure = Figure((axes,), title="")
    view = FigureView(model.figure)
    assert model.title == model.axes.title == expected_titles[0]
    assert figure.axes[0].title == expected_titles[0]
    assert view.figure.axes[0].get_title() == expected_titles[0]
    model.ys.append("det+1")
    assert model.title == model.axes.title == expected_titles[1]
    assert figure.axes[0].title == expected_titles[1]
    assert view.figure.axes[0].get_title() == expected_titles[1]
    model.title = test_titles[1]
    assert model.title == model.axes.title == expected_titles[2]
    assert figure.axes[0].title == expected_titles[2]
    assert view.figure.axes[0].get_title() == expected_titles[2]
    model.ys.append("det+2")
    assert model.title == model.axes.title == expected_titles[3]
    assert figure.axes[0].title == expected_titles[3]
    assert view.figure.axes[0].get_title() == expected_titles[3]

    view.close()



"""
An example integration QtFigures into an "existing" Qt application. Run like:

python -m bluesky_widgets.examples.qt_app_integration
"""
from qtpy.QtWidgets import QApplication, QVBoxLayout, QLabel, QMainWindow, QWidget


def main():
    # First, some boilerplate to make a super-minimal Qt application that we want
    # to add some bluesky-widgets components into.
    app = QApplication(["Some App"])
    window = QMainWindow()
    central_widget = QWidget(window)
    window.setCentralWidget(central_widget)
    central_widget.setLayout(QVBoxLayout())
    central_widget.layout().addWidget(QLabel("This is part of the 'original' app."))
    window.show()

    # *** INTEGRATION WITH BLUESKY-WIDGETS STARTS HERE. ***

    # Ensure that any background workers started by bluesky-widgets stop
    # gracefully when the application closes.
    from bluesky_widgets.qt.threading import wait_for_workers_to_quit

    app.aboutToQuit.connect(wait_for_workers_to_quit)

    # Model a list of figures.
    # This will generate line plot automatically based on the structure (shape)
    # of the data and its hints. Other models could be used for more explicit
    # control of what gets plotted. See the examples in
    # http://blueskyproject.io/bluesky-widgets/reference.html#plot-builders
    from bluesky_widgets.models.auto_plot_builders import AutoLines

    model = AutoLines(max_runs=3)

    # Feed it data from the RunEngine. In actual practice, the RunEngine should
    # be in a separate process and we should be receiving these documents
    # over a network via publish--subscribe. See
    # bluesky_widgets.examples.advanced.qt_viewer_with_search for an example of
    # that. Here, we keep it simple.
    from bluesky_widgets.utils.streaming import stream_documents_into_runs
    from bluesky import RunEngine

    RE = RunEngine()
    RE.subscribe(stream_documents_into_runs(model.add_run))

    # Add a tabbed pane of figures to the app.
    from bluesky_widgets.qt.figures import QtFigures

    view = QtFigures(model.figures)  # view is a QWidget
    central_widget.layout().addWidget(view)

    # When the model receives data or is otherwise updated, any changes to
    # model.figures will be reflected in changes to the view.

    # Just for this example, generate some data before starting this app.
    # Again, in practice, this should happen in a separate process and send
    # the results over a network.

    from bluesky.plans import scan
    from ophyd.sim import motor, det

    def plan():
        for i in range(1, 5):
            yield from scan([det], motor, -1, 1, 1 + 2 * i)

    RE(plan())

    # *** INTEGRATION WITH BLUESKY-WIDGETS ENDS HERE. ***

    # Run the app.
    app.exec_()


if __name__ == "__main__":
    main()



import ast
import collections
import contextlib
import inspect

import numpy

from ..utils.list import EventedList
from ..utils.event import EmitterGroup, Event


class RunList(EventedList):
    """
    A list of BlueskyRuns.
    """

    __slots__ = ()

    def __contains__(self, run):
        uid = run.metadata["start"]["uid"]
        for run_ in self:
            if run_.metadata["start"]["uid"] == uid:
                return True
        else:
            return False


def run_is_completed(run):
    "True is Run is completed and no further updates are coming."
    return run.metadata["stop"] is not None


def run_is_live(run):
    "True if Run is 'live' (observable) based on a streaming source, not at rest."
    return hasattr(run, "events")


def run_is_live_and_not_completed(run):
    "True if Run is 'live' (observable) and not yet complete."
    return run_is_live(run) and (not run_is_completed(run))


@contextlib.contextmanager
def lock_if_live(run):
    """
    Lock to prevent new data from being added, if this is a 'live' BlueskyRun.

    If it is a BlueskyRun backed by data at rest (i.e. from databroker) do
    nothing.
    """
    if run_is_live(run):
        with run.write_lock:
            yield
    else:
        yield


# Make numpy functions accessible as (for example) log, np.log, and numpy.log.
_base_namespace = {"numpy": numpy, "np": numpy}
_base_namespace.update({name: getattr(numpy, name) for name in numpy.__all__})


def construct_namespace(run, stream_names):
    """
    Put the contents of a run into a namespace to lookup in or ``eval`` expressions in.

    This is used by the plot builders to support usages like

    >>> model = Lines("-motor", ["log(det)"])

    The words available in these expressions include:

    * The ``BlueskyRun`` itself, as ``"run"``, from which any data or metadata
      can be obtained
    * All the streams, with the data accessible as items in a dict, as in
      ``"primary['It']"`` or ``"baseline['motor']"``
    * The columns in the streams given by stream_names, as in ``"I0"``. If a
      column name appears in multiple streams, the streams earlier in the list
      get precedence.
    * All functions in numpy. They can be spelled as ``log``, ``np.log``, or
      ``numpy.log``

    In the event of name collisions, items higher in the list above will get
    precedence.

    Parameters
    ----------
    run : BlueskyRun
    stream_names : List[String]

    Returns
    -------
    namespace : Dict
    """
    namespace = dict(_base_namespace)  # shallow copy
    with lock_if_live(run):
        run_start_time = run.metadata["start"]["time"]
        # Add columns from streams in stream_names. Earlier entries will get
        # precedence.
        for stream_name in reversed(stream_names):
            ds = run[stream_name].to_dask()
            namespace.update({column: ds[column] for column in ds})
            namespace.update({column: ds[column] for column in ds.coords})
        if "time" in namespace:
            namespace["time"] = namespace["time"] - run_start_time
        namespace.update({stream_name: run[stream_name].to_dask() for stream_name in stream_names})
        for stream_name in stream_names:
            namespace[stream_name]["time"] = namespace[stream_name]["time"] - run_start_time
    namespace.update({"run": run})
    return namespace


class BadExpression(Exception):
    pass


def call_or_eval(mapping, run, stream_names, namespace=None):
    """
    Given a mix of callables and string expressions, call or eval them.

    Parameters
    ----------
    mapping : Dict[String, String | Callable]
        Each item must be a stream name, field name, a valid Python
        expression, or a callable. The signature of the callable may include
        any valid Python identifiers provideed by :func:`construct_namespace`
        or the user-provided namespace parmeter below. See examples.
    run : BlueskyRun
    stream_names : List[String]
    namespace : Dict, optional

    Returns
    -------
    results : Dict[String, Any]

    Raises
    ------
    ValueError
        If input is not String or Callable
    BadExpression
        If input is String and eval(...) raises an error

    Examples
    --------

    A function can have access to the whole BlueskyRun.

    >>> def f(run):
    ...     ds = run.primary.read()
    ...     return (ds["a"] - ds["b"]) / (ds["a"] + ds["b"])
    ...
    >>> call_or_eval({"x": f}, run, ["primary"])

    But, it also provides a more "magical" option in support of brevity.
    The signature may include parameters with the names streams or fields. The
    names in the signature are significant and will determine what parameters
    the function is called with.

    >>> def f(a, b):
    ...     return (a - b) / (a + b)
    ...
    >>> call_or_eval({"x": f}, run, ["primary"])

    Equivalently, as a lambda function:
    >>> call_or_eval({"f": lambda a, b: (a - b) / (a + b)}, run, ["primary"])
    """
    with lock_if_live(run):
        namespace_ = construct_namespace(run, stream_names)
        # Overlay user-provided namespace.
        namespace_.update(namespace or {})
        del namespace  # Avoid conflating namespace and namespace_ below.

        return {key: call_or_eval_one(item, namespace_) for key, item in mapping.items()}


def call_or_eval_one(item, namespace):
    """
    Given a mix of callables and string expressions, call or eval them.

    Parameters
    ----------
    item : String | Callable
        Each item must be a stream name, field name, a valid Python
        expression, or a callable. The signature of the callable may include
        any valid Python identifiers provideed in the namespace.

    namespace : Dict
        The namespace that the item is evaluated against.

    Returns
    -------
    result : Any

    Raises
    ------
    ValueError
        If input is not String or Callable
    BadExpression
        If input is String and eval(...) raises an error

    """
    # If it is a callable, call it.
    if callable(item):
        # Inspect the callable's signature. For each parameter, find an
        # item in our namespace with a matching name. This is similar
        # to the "magic" of pytest fixtures.
        parameters = inspect.signature(item).parameters
        kwargs = {}
        for name, parameter in parameters.items():
            try:
                kwargs[name] = namespace[name]
            except KeyError:
                if parameter.default is parameter.empty:
                    raise ValueError(f"Cannot find match for parameter {name}")
                # Otherwise, it's an optional parameter, so skip it.
        return item(**kwargs)
    elif isinstance(item, str):
        # If it is a key in our namespace, look it up.
        try:
            # This handles field or stream names that are not valid
            # Python identifiers (e.g. ones with spaces in them).
            return namespace[item]
        except KeyError:
            pass
        # Check whether it is valid Python syntax.
        try:
            ast.parse(item)
        except SyntaxError as err:
            raise ValueError(f"Could find {item!r} in namespace or parse it as a Python expression.") from err
        # Try to evaluate it as a Python expression in the namespace.
        try:
            return eval(item, namespace)
        except Exception as err:
            raise ValueError(f"Could find {item!r} in namespace or evaluate it.") from err
    else:
        raise ValueError(f"expected callable or string, received {item!r} of type {type(item).__name__}")


def auto_label(callable_or_expr):
    """
    Given a callable or a string, extract a name for labeling axes.

    Parameters
    ----------
    callable_or_expr : String | Callable

    Returns
    -------
    label : String
    """
    if callable(callable_or_expr):
        return getattr(callable_or_expr, "__name__", repr(callable_or_expr))
    elif isinstance(callable_or_expr, str):
        return callable_or_expr
    else:
        raise ValueError(
            f"expected callable or string, received {callable_or_expr!r} of "
            f"type {type(callable_or_expr).__name__}"
        )


class RunManager:
    """
    Keep a RunList with a maximum number of Runs, plus any 'pinned' Runs.

    This is used internally as a helper class by Lines, Images, and others.
    This tracks the relationship between Runs and Artists and ensures correct
    cleanup when a Run is removed.
    """

    def __init__(self, max_runs, needs_streams):
        self._max_runs = int(max_runs)
        self._needs_streams = tuple(needs_streams)
        self.runs = RunList()
        self._pinned = set()
        # Maps Run (uid) to set of ArtistSpec.
        self._runs_to_artists = collections.defaultdict(list)

        self.runs.events.added.connect(self._on_run_added)
        self.runs.events.removed.connect(self._on_run_removed)
        self.events = EmitterGroup(source=self, run_ready=Event)

    def add_run(self, run, *, pinned=False):
        """
        Add a Run.

        Parameters
        ----------
        run : BlueskyRun
        pinned : Boolean
            If True, retain this Run until it is removed by the user.
        """
        if pinned:
            self._pinned.add(run.metadata["start"]["uid"])
        self.runs.append(run)

    def discard_run(self, run):
        """
        Discard a Run, including any pinned and unpinned.

        If the Run is not present, this will return silently.

        Parameters
        ----------
        run : BlueskyRun
        """
        if run in self.runs:
            self.runs.remove(run)

    def track_artist(self, artist, runs):
        """
        Track an Artist.

        This ensures it will be removed when the associated Run is removed.

        Parameters
        ----------
        artist : ArtistSpec
        runs : List[BlueskyRun]
        """
        # TODO Someday we will need aritsts that represent data from *multiple*
        # runs, and then we will need to rethink the expected API of artist
        # (.run -> .runs?) and the cache management here. But that would be a
        # widereaching change, so we'll stay within the framework as it is
        # today.
        if len(runs) != 1:
            raise NotImplementedError("We current assume a 1:1 association of aritsts and runs.")
        (run,) = runs
        run_uid = run.metadata["start"]["uid"]
        self._runs_to_artists[run_uid].append(artist)

    def _cull_runs(self):
        "Remove Runs from the beginning of self.runs to keep the length <= max_runs."
        i = 0
        while len(self.runs) > self.max_runs + len(self._pinned):
            while self.runs[i].metadata["start"]["uid"] in self._pinned:
                i += 1
            self.runs.pop(i)

    def _on_run_added(self, event):
        """
        When a new Run is added, mark it as ready or listen for it to become ready.

        By "ready" we mean, it has all the streams it needs to be drawn.
        """
        self._cull_runs()
        run = event.item
        if run_is_live_and_not_completed(run):
            # If the stream of interest is defined already, plot now.
            if set(self.needs_streams).issubset(set(list(run))):
                self.events.run_ready(run=run)
            else:
                # Otherwise, connect a callback to run when the stream of interest arrives.
                run.events.new_stream.connect(self._on_new_stream)
        else:
            if set(self.needs_streams).issubset(set(list(run))):
                self.events.run_ready(run=run)

    def _on_run_removed(self, event):
        "Remove any extant artists if its corresponding Run is removed."
        run_uid = event.item.metadata["start"]["uid"]
        self._pinned.discard(run_uid)
        for artist in self._runs_to_artists.pop(run_uid):
            artist.axes.discard(artist)

    def _on_new_stream(self, event):
        "When an unready Run get a new stream, check it if is now ready."
        if set(self.needs_streams).issubset(set(list(event.run))):
            self.events.run_ready(run=event.run)
            event.run.events.new_stream.disconnect(self._on_new_stream)

    @property
    def max_runs(self):
        return self._max_runs

    @max_runs.setter
    def max_runs(self, value):
        self._max_runs = value
        self._cull_runs()

    @property
    def pinned(self):
        return frozenset(self._pinned)

    @property
    def needs_streams(self):
        return self._needs_streams



import logging
import time

from bluesky import RunEngine
from bluesky.plans import count

from bluesky_widgets.models.plot_builders import Lines
from bluesky_widgets.utils.streaming import stream_documents_into_runs

from bluesky_widgets.qt.kafka_dispatcher import QtRemoteDispatcher


test_logger = logging.getLogger("bluesky_widgets.qt.tests")


def test_publisher_and_qt_remote_dispatcher(
    kafka_bootstrap_servers, temporary_topics, publisher_factory, qapp, hw
):
    """Test publishing and dispatching bluesky documents in Kafka messages.
    Messages will be "dispatched" by a `bluesky_kafka.RemoteDispatcher`.
    Parameters
    ----------
    kafka_bootstrap_servers: str (pytest fixture)
        comma-delimited string of Kafka broker host:port, for example "localhost:9092"
    temporary_topics: context manager (pytest fixture)
        creates and cleans up temporary Kafka topics for testing
    publisher_factory: pytest fixture
        fixture-as-a-factory for creating Publishers
    qapp: pytest-qt fixture
        needed to force processing of Qt events
    hw: pytest fixture
        ophyd simulated hardware objects
    """

    with temporary_topics(topics=["test.qt.remote.dispatcher"]) as (topic,):

        bluesky_publisher = publisher_factory(
            topic=topic,
            key=f"{topic}.key",
            flush_on_stop_doc=True,
        )

        # trying to test _waiting_for_start
        bluesky_publisher("descriptor", {})
        bluesky_publisher("event", {})
        # bluesky_publisher("stop", {})

        published_bluesky_documents = []

        # this function will store all documents
        # published by the RunEngine in a list
        def store_published_document(name, document):
            published_bluesky_documents.append((name, document))

        RE = RunEngine()
        RE.subscribe(bluesky_publisher)
        RE.subscribe(store_published_document)

        RE(count([hw.det]))

        # it is assumed that RE(count()) will produce four
        # documents: start, descriptor, event, stop
        assert len(published_bluesky_documents) == 4

        lines_model = Lines(x="time", ys=["det"])
        assert len(lines_model.runs) == 0

        qt_remote_dispatcher = QtRemoteDispatcher(
            topics=[topic],
            bootstrap_servers=kafka_bootstrap_servers,
            group_id=f"{topic}.consumer.group",
            consumer_config={
                # this consumer is intended to read messages that
                # were published before it starts, so it is necessary
                # to specify "earliest" here
                "auto.offset.reset": "earliest",
            },
            polling_duration=1.0,
        )

        qt_remote_dispatcher.subscribe(stream_documents_into_runs(lines_model.add_run))

        dispatched_bluesky_documents = []

        # the QtRemoteDispatcher will use this function
        # to build a list of documents delivered by Kafka
        def store_dispatched_document(name, document):
            test_logger.debug("store_dispatched_document name='%s'", name)
            dispatched_bluesky_documents.append((name, document))

        qt_remote_dispatcher.subscribe(store_dispatched_document)

        # this function will be given to QtRemoteDispatcher.start()
        # it returns False to end the polling loop
        # as soon as it sees one stop document
        def until_first_stop_document():
            dispatched_bluesky_document_names = [name for name, _ in dispatched_bluesky_documents]
            test_logger.debug("until_first_stop_document %s", dispatched_bluesky_document_names)
            if "stop" in dispatched_bluesky_document_names:
                return False
            else:
                return True

        # start() will return when 'until_first_stop_document' returns False
        qt_remote_dispatcher.start(
            continue_polling=until_first_stop_document,
        )

        while len(dispatched_bluesky_documents) < 3:
            # waiting for all Kafka messages to land
            test_logger.debug("processing Qt Events")
            qapp.processEvents()

            time.sleep(1.0)
            test_logger.debug(
                "waiting for all Kafka messages, %d so far",
                len(dispatched_bluesky_documents),
            )

        assert len(published_bluesky_documents) == len(dispatched_bluesky_documents)

        assert len(lines_model.runs) == 1



from bluesky_live.run_builder import RunBuilder
import pytest
import numpy

from ..plot_builders import RasteredImages
from ..plot_specs import Axes, Image, Figure


@pytest.fixture
def non_snaking_run():
    # Test data
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, False)}
    with RunBuilder(md) as builder:
        builder.add_stream("primary", data={"ccd": [1, 2, 3, 4], "x": [0, 1, 0, 1], "y": [0, 0, 1, 1]})
    run = builder.get_run()
    return run


@pytest.fixture
def snaking_run():
    # Test data
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, True)}
    with RunBuilder(md) as builder:
        builder.add_stream("primary", data={"ccd": [1, 2, 3, 4], "x": [0, 1, 1, 0], "y": [0, 0, 1, 1]})
    run = builder.get_run()
    return run


def test_rastered_image(non_snaking_run, FigureView):
    "Test RasteredImages with a 2D array."
    run = non_snaking_run
    model = RasteredImages("ccd", shape=(2, 2))
    view = FigureView(model.figure)
    assert not model.figure.axes[0].artists
    model.add_run(run)
    assert model.figure.axes[0].artists
    view.close()


def test_x_y_positive_change_x_y_limits(non_snaking_run, FigureView):
    "Test x_positive and y_positive change x_limits and y_limits"
    run = non_snaking_run
    model = RasteredImages("ccd", shape=(2, 2), x_positive="left", y_positive="down")
    view = FigureView(model.figure)
    model.add_run(run)
    expected_x_lims = expected_y_lims = (1.5, -0.5)
    assert model.axes.x_limits == expected_x_lims
    assert model.axes.y_limits == expected_y_lims
    model.x_positive = "right"
    model.y_positive = "up"
    expected_x_lims = expected_y_lims = (-0.5, 1.5)
    assert model.axes.x_limits == expected_x_lims
    assert model.axes.y_limits == expected_y_lims
    view.close()


def test_x_y_limits_change_x_y_positive(non_snaking_run, FigureView):
    "Test x_limits and y_limits change x_positive and y_positive"
    run = non_snaking_run
    axes = Axes(x_limits=(1.5, -0.5), y_limits=(1.5, -0.5))
    Figure((axes,), title="")
    model = RasteredImages("ccd", shape=(2, 2), axes=axes)
    view = FigureView(model.figure)
    model.add_run(run)
    assert model.x_positive == "left"
    assert model.y_positive == "down"
    model.axes.x_limits = model.axes.y_limits = (-0.5, 1.5)
    assert model.x_positive == "right"
    assert model.y_positive == "up"
    view.close()


def test_extent_setting(non_snaking_run, FigureView):
    """
    Test if ``RasteredImages.extent`` is properly set.
    """
    run = non_snaking_run
    axes = Axes()
    Figure((axes,), title="")
    extent1 = (0, 1, 4, 5)
    model = RasteredImages("ccd", shape=(2, 2), axes=axes, extent=extent1)
    view = FigureView(model.figure)
    model.add_run(run)

    def check_artists(ext):
        image_found = False
        for artist in axes.artists:
            if isinstance(artist, Image):
                assert "extent" in artist.style
                assert artist.style["extent"] == ext
                image_found = True
        if not image_found:
            assert "No Image found in the list of artists"

    assert model.extent == extent1
    check_artists(extent1)

    extent2 = (1, 2, 6, 7)
    model.extent = extent2
    assert model.extent == extent2
    check_artists(extent2)

    view.close()


def test_show_colorbar_setting(non_snaking_run, FigureView):
    """
    Test if ``RasteredImages.show_colorbar`` is properly set.
    """
    run = non_snaking_run
    axes = Axes()
    Figure((axes,), title="")
    model = RasteredImages("ccd", shape=(2, 2), axes=axes, show_colorbar=True)
    view = FigureView(model.figure)
    model.add_run(run)

    def check_artists(value):
        image_found = False
        for artist in axes.artists:
            if isinstance(artist, Image):
                assert "show_colorbar" in artist.style
                assert artist.style["show_colorbar"] == value
                image_found = True
        if not image_found:
            assert "No Image found in the list of artists"

    assert model.show_colorbar is True
    check_artists(True)

    # Check if setting the property changes the style of the artists in the next image
    model.show_colorbar = False
    model.add_run(run)
    assert model.show_colorbar is False
    check_artists(False)

    view.close()


def test_non_snaking_image_data(non_snaking_run, FigureView):
    run = non_snaking_run
    model = RasteredImages("ccd", shape=(2, 2))
    model.add_run(run)
    view = FigureView(model.figure)
    actual_data = model.figure.axes[0].artists[0].update()["array"]
    expected_data = [[1, 2], [3, 4]]
    assert numpy.array_equal(actual_data, expected_data)
    view.close()


def test_snaking_image_data(snaking_run, FigureView):
    run = snaking_run
    model = RasteredImages("ccd", shape=(2, 2))
    view = FigureView(model.figure)
    model.add_run(run)
    actual_data = model.figure.axes[0].artists[0].update()["array"]
    expected_data = [[1, 2], [4, 3]]
    assert numpy.array_equal(actual_data, expected_data)
    view.close()


def test_non_snaking_image_data_positions(FigureView):
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, False)}
    model = RasteredImages("ccd", shape=(2, 2))
    view = FigureView(model.figure)
    with RunBuilder(md) as builder:
        ccd = iter([1, 2, 3, 4])
        x = iter([0, 1, 0, 1])
        y = iter([0, 0, 1, 1])
        run = builder.get_run()
        model.add_run(run)
        # First data point
        builder.add_stream("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, numpy.nan], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Second point
        builder.add_data("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, 2], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Third point
        builder.add_data("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, 2], [3, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Fourth point
        builder.add_data("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, 2], [3, 4]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
    view.close()


def test_snaking_image_data_positions(FigureView):
    md = {"motors": ["y", "x"], "shape": [2, 2], "snaking": (False, True)}
    model = RasteredImages("ccd", shape=(2, 2))
    view = FigureView(model.figure)
    with RunBuilder(md) as builder:
        ccd = iter([1, 2, 3, 4])
        x = iter([0, 1, 1, 0])
        y = iter([0, 0, 1, 1])
        run = builder.get_run()
        model.add_run(run)
        # First data point
        builder.add_stream("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, numpy.nan], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Second point
        builder.add_data("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, 2], [numpy.nan, numpy.nan]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Third point
        builder.add_data("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, 2], [numpy.nan, 3]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
        # Fourth point
        builder.add_data("primary", data={"ccd": [next(ccd)], "x": [next(x)], "y": [next(y)]})
        actual_data = model.figure.axes[0].artists[0].update()["array"]
        expected_data = [[1, 2], [4, 3]]
        assert numpy.array_equal(actual_data, expected_data, equal_nan=True)
    view.close()


def test_figure_set_after_instantiation():
    axes = Axes()
    model = RasteredImages("ccd", shape=(2, 2), axes=axes)
    assert model.figure is None
    figure = Figure((axes,), title="")
    assert model.figure is figure



from datetime import datetime, timedelta
import pytest
from ...examples.qt_search import ExampleApp


@pytest.fixture(scope="function")
def make_test_app(qtbot, request):
    apps = []

    def actual_factory(*model_args, **model_kwargs):
        model_kwargs["show"] = model_kwargs.pop("show", request.config.getoption("--show-window"))
        app = ExampleApp(*model_args, **model_kwargs)
        apps.append(app)
        return app

    yield actual_factory

    for app in apps:
        app.close()


def test_app(make_test_app):
    "An integration test"
    app = make_test_app()
    app.searches[0].input.since = datetime(1980, 2, 2)
    app.searches[0].input.since = datetime(1985, 11, 15)
    app.searches[0].input.since = timedelta(days=-1)



import napari

from bluesky_widgets.examples.utils.add_search_mixin import extract_results_row_from_run
from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog
from bluesky_widgets.examples.utils.get_run_images import generate_thumbnail
from bluesky_widgets.models.search import Search, SearchList
from bluesky_widgets.utils.event import Event
from qtpy.QtWidgets import QWidget, QPushButton, QVBoxLayout
from bluesky_widgets.qt.search import QtSearches


class SearchListWithButton(SearchList):
    """
    A SearchList model with a method to handle a click event.

    If we were to add a method of this class called on_view_images,
    then it would automatically receive view_images events without
    needing to use view_images.connect.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Register an event emitter for process events.
        # The kwarg view_images defines the name of the events that are emitted.
        # Event is the event type.
        self.events.add(view_images=Event)


class QtSearchListWithButton(QWidget):
    """
    A view for SearchListWithButton.

    Combines the QtSearches widget with a button that processes the selected Runs.
    """

    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(QtSearches(model))

        # Add a button that does something with the currently-selected Runs
        # when you click it.
        go_button = QPushButton("Process Selected Runs")
        layout.addWidget(go_button)

        # Register a callback (slot) for the button qt click signal.
        go_button.clicked.connect(self.on_click)

    def on_click(self):
        """
        Receive the qt signal, and emmit a view_images event.

        Add whatever we want to the event by passing it as a kwarg, in this case the images kwarg.
        Can access images in the callback by accessing Event.images.
        """
        thumbnails = [generate_thumbnail(run) for _, run in self.model.active.selection_as_catalog.items()]
        self.model.events.view_images(images=thumbnails)


class NapariDatabroker(napari.Viewer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.searches = SearchListWithButton()

        # Register the process event callback.
        self.searches.events.view_images.connect(self.load_images_from_run)

    def add_search(self, catalog, columns=extract_results_row_from_run):
        """
        Add a new Search form.
        """
        headings = (
            "Unique ID",
            "Transient Scan ID",
            "Plan Name",
            "Start Time",
            "Duration",
            "Exit Status",
        )

        search = Search(catalog, columns=(headings, extract_results_row_from_run))
        self.searches.append(search)

    @property
    def active_search(self):
        """
        Convenience for accessing the currently-active Search form.
        """
        return self.searches.active

    def load_images_from_run(self, view_images):
        """
        Handle a view_images event.
        """
        for thumbnail in view_images.images:
            self.add_image(thumbnail)


with napari.gui_qt():
    viewer = NapariDatabroker()
    viewer.grid.enabled = True
    viewer.window.add_dock_widget(QtSearchListWithButton(viewer.searches), area="right")

    # Initialize with a two search tabs: one with some generated example data...
    viewer.add_search(get_catalog())
    # ...and one listing any and all catalogs discovered on the system.
    from databroker import catalog

    viewer.add_search(catalog)



from bluesky_live.run_builder import build_simple_run, RunBuilder
import numpy
import xarray
import pytest

from ..utils import call_or_eval, construct_namespace


def test_namespace():
    "Test the contents of a namespace for eval-ing expressions with a run."
    run = build_simple_run({"motor": [1, 2], "det": [10, 20]})
    namespace = construct_namespace(run, ["primary"])

    # Test entities from run....
    # the run itself
    assert eval("run", namespace) is run
    # a stream
    assert numpy.array_equal(eval("primary['motor']", namespace), numpy.array([1, 2]))
    # a field in the 'primary' stream
    assert numpy.array_equal(eval("motor", namespace), numpy.array([1, 2]))
    # numpy, three different ways
    expected = 3 + numpy.log(numpy.array([1, 2]))
    assert numpy.array_equal(eval("3 + log(motor)", namespace), expected)
    assert numpy.array_equal(eval("3 + np.log(motor)", namespace), expected)
    assert numpy.array_equal(eval("3 + numpy.log(motor)", namespace), expected)


def test_collision_with_the_name_run():
    "AHHH everything is named 'run'! The BlueskyRun should take precedence."
    with RunBuilder() as builder:
        builder.add_stream("run", data={"run": [1, 2]})
    run = builder.get_run()
    namespace = construct_namespace(run, ["run"])
    assert eval("run", namespace) is run


def test_collision_of_stream_name_and_field_name():
    "If there is a field named 'primary', the stream should take precedence."
    run = build_simple_run({"primary": [1, 2], "det": [10, 20]})
    namespace = construct_namespace(run, ["primary"])
    assert isinstance(eval("primary", namespace), xarray.Dataset)
    assert isinstance(eval("det", namespace), xarray.DataArray)


def test_collision_of_fields_across_streams():
    "The field in the stream listed first in needs_streams should take precedence."
    with RunBuilder() as builder:
        # Two streams, each with a field named "a".
        builder.add_stream("primary", data={"a": [1, 2, 3, 2, 1]})
        builder.add_stream("baseline", data={"a": [1, 1]})
    run = builder.get_run()
    namespace1 = construct_namespace(run, ["primary", "baseline"])
    # We should get run.primary.read()["a"] here.
    assert len(eval("a", namespace1)) == 5
    namespace2 = construct_namespace(run, ["baseline", "primary"])
    # We should get run.secondary.read()["a"] here.
    assert len(eval("a", namespace2)) == 2


def test_stream_omitted():
    "A stream not in `needs_stream` should have it and its fields omitted."
    with RunBuilder() as builder:
        # Two streams, each with a field named "a".
        builder.add_stream("primary", data={"a": [1, 2, 3, 2, 1]})
        builder.add_stream("baseline", data={"b": [1, 1]})
    run = builder.get_run()
    namespace = construct_namespace(run, ["baseline"])
    assert "a" not in namespace
    assert "primary" not in namespace
    assert "baseline" in namespace
    assert "b" in namespace


def test_names_with_spaces():
    "Names with spaces cannot be eval-ed, but they can be looked up."
    with RunBuilder() as builder:
        # Two streams, each with a field named "a".
        builder.add_stream("some stream", data={"some field": [1, 2, 3, 2, 1]})
    run = builder.get_run()
    namespace = construct_namespace(run, ["some stream"])
    assert "some stream" in namespace
    assert "some field" in namespace


def test_call_or_eval_errors():
    "Test that various failrue modes raise expected errors."
    run = build_simple_run({"motor": [1, 2], "det": [10, 20]})
    with pytest.raises(ValueError, match=".*callable or string.*"):
        call_or_eval({"x": 1}, run, ["primary"])
    with pytest.raises(ValueError, match=".*parse.*"):
        call_or_eval({"x": "invalid***syntax"}, run, ["primary"])
    with pytest.raises(ValueError, match=".*evaluate.*"):
        call_or_eval({"x": "missing_key"}, run, ["primary"])


def test_call_or_eval_with_user_namespace():
    "Test that user-injected items in the namespace are found."
    run = build_simple_run({"motor": [1, 2], "det": [10, 20]})
    thing = object()
    result = call_or_eval({"x": "thing"}, run, [], namespace={"thing": thing})
    assert result["x"] is thing


def test_call_or_eval_magical_signature_inspection():
    "Test magical signature inspection."
    run = build_simple_run({"motor": [1, 2], "det": [10, 20]})

    def func1(motor, det):
        "Access fields by name."
        return motor + det

    result = call_or_eval({"x": func1}, run, ["primary"])
    assert numpy.array_equal(result["x"], [11, 22])

    def func2(primary):
        "Access a stream by name."
        return primary["motor"]

    result = call_or_eval({"x": func2}, run, ["primary"])
    assert numpy.array_equal(result["x"], [1, 2])

    def func3(does_not_exist):
        "Test a missing variable."
        ...

    with pytest.raises(ValueError, match="Cannot find match for.*"):
        call_or_eval({"x": func3}, run, ["primary"])

    def func4(thing):
        "Test an item in the user-provided namespace."
        return thing

    thing = object()
    result = call_or_eval({"x": func4}, run, [], namespace={"thing": thing})
    assert result["x"] is thing



from bluesky_live.run_builder import RunBuilder
from bluesky_live.event import CallbackException
import pytest

from ...models.plot_specs import (
    AxesAlreadySet,
    Figure,
    FigureList,
    Axes,
    Line,
    Image,
)


# Generate example data.
with RunBuilder() as builder:
    builder.add_stream("primary", data={"a": [1, 2, 3], "b": [1, 4, 9]})
run = builder.get_run()


def transform(run):
    ds = run.primary.read()
    return {"x": ds["a"], "y": ds["b"]}


def func(run):
    line = Line.from_run(transform, run, "label")
    axes = Axes(artists=[line], x_label="a", y_label="b", title="axes title")
    figure = Figure((axes,), title="figure title")
    return figure


def test_figure(FigureView):
    "Basic test: create a FigureView."
    figure = func(run)
    FigureView(figure)


def test_figures(FigureViews):
    "Basic test: create FigureViews."
    figure = func(run)
    another_figure = func(run)
    figures = FigureList([figure, another_figure])
    FigureViews(figures)


def test_figure_title_syncing(FigureView):
    model = func(run)
    view = FigureView(model)
    initial = model.title
    assert view.figure._suptitle.get_text() == initial
    expected = "figure title changed"
    model.title = expected
    assert model.title == expected
    # Is there no public matplotlib API for this?
    # https://stackoverflow.com/a/48917679
    assert view.figure._suptitle.get_text() == expected


def test_short_title_syncing(FigureViews, request):
    QtFigures = pytest.importorskip("bluesky_widgets.qt.figures.QtFigures")
    if request.getfixturevalue("FigureViews") is not QtFigures:
        pytest.skip("This tests details of the QtFigures view.")
    model = func(run)
    figures = FigureList([model])
    view = FigureViews(figures)
    actual_title = view.figures[model.uuid].figure._suptitle.get_text()
    assert view.tabText(0) == actual_title
    expected_short_title = "new short title"
    model.short_title = expected_short_title
    assert model.short_title == expected_short_title
    actual_title = view.figures[model.uuid].figure._suptitle.get_text()
    assert view.tabText(0) == expected_short_title
    assert actual_title == model.title
    expected_title = "new title"
    model.title = expected_title
    assert view.tabText(0) == model.short_title
    model.short_title = None
    assert view.tabText(0) == expected_title


def test_non_null_short_title_syncing(FigureViews, request):
    QtFigures = pytest.importorskip("bluesky_widgets.qt.figures.QtFigures")
    if request.getfixturevalue("FigureViews") is not QtFigures:
        pytest.skip("This tests details of the QtFigures view.")
    model = func(run)
    model.short_title = "short title"
    figures = FigureList([model])
    view = FigureViews(figures)
    actual_title = view.figures[model.uuid].figure._suptitle.get_text()
    assert view.tabText(0) == model.short_title
    assert actual_title == model.title


@pytest.mark.parametrize(
    ("model_property", "mpl_method"),
    [("title", "get_title"), ("x_label", "get_xlabel"), ("y_label", "get_ylabel")],
)
def test_axes_syncing(FigureView, model_property, mpl_method):
    model = func(run)
    view = FigureView(model)
    initial = getattr(model.axes[0], model_property)
    assert getattr(view.figure.axes[0], mpl_method)() == initial
    expected = "axes title changed"
    setattr(model.axes[0], model_property, expected)
    assert getattr(model.axes[0], model_property) == expected
    assert getattr(view.figure.axes[0], mpl_method)() == expected


def test_axes_set_figure():
    "Adding axes to a figure sets their figure."
    axes = Axes()
    assert axes.figure is None
    figure = Figure((axes,), title="figure title")
    assert axes.figure is figure
    # Once axes belong to a figure, they cannot belong to another figure.
    with pytest.raises(RuntimeError):
        Figure((axes,), title="figure title")

    with pytest.raises(AttributeError):
        figure.axes = (axes,)  # not settable


artist_set_axes_params = pytest.mark.parametrize(
    "artist_factory",
    # These are factories because each artist can only be assigned to Axes once
    # in its lifecycle. For each test that these params are used in, we need a
    # fresh instance.
    [
        lambda: Line.from_run(transform, run, "label"),
        lambda: Image.from_run(transform, run, "label"),
    ],
    ids=["lines", "images"],
)


@artist_set_axes_params
def test_artist_set_axes_at_init(artist_factory):
    "Adding an artist to axes at init time sets its axes."
    artist = artist_factory()
    axes = Axes(artists=[artist])
    assert artist in axes.artists
    assert artist in axes.by_uuid.values()
    assert artist.axes is axes

    # Once line belong to a axes, it cannot belong to another axes.
    with pytest.raises(CallbackException) as exc_info:
        Axes(artists=[artist])
    exc = exc_info.value
    assert hasattr(exc, "__cause__") and isinstance(exc.__cause__, AxesAlreadySet)


@artist_set_axes_params
def test_artist_set_axes_after_init(artist_factory):
    "Adding an artist to axes after init time sets its axes."
    artist = artist_factory()
    axes = Axes()
    axes.artists.append(artist)
    assert artist in axes.artists
    assert artist in axes.by_uuid.values()
    assert artist.axes is axes

    # Once line belong to a axes, it cannot belong to another axes.
    with pytest.raises(CallbackException) as exc_info:
        Axes(artists=[artist])
    exc = exc_info.value
    assert hasattr(exc, "__cause__") and isinstance(exc.__cause__, AxesAlreadySet)



"""
Extendeding and supplementing the widgets import bluesky-widgets
"""
from bluesky_widgets.qt.run_engine_client import (
    QtReEnvironmentControls,
    QtReManagerConnection,
    QtReQueueControls,
    QtReExecutionControls,
    QtReStatusMonitor,
    QtRePlanQueue,
    QtRePlanHistory,
    QtReRunningPlan,
    QtRePlanEditor,
    QtReConsoleMonitor,
)
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QTabWidget,
    QFrame,
    QSplitter,
)
from qtpy.QtCore import Qt


class QtOrganizeQueueWidgets(QSplitter):
    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model

        self.setOrientation(Qt.Vertical)

        self._frame_1 = QFrame(self)
        self._frame_2 = QFrame(self)
        self._frame_3 = QFrame(self)
        self._frame_4 = QFrame(self)

        self.addWidget(self._frame_1)
        self.addWidget(self._frame_2)
        self.addWidget(self._frame_3)
        self.addWidget(self._frame_4)

        self._running_plan = QtReRunningPlan(model)
        self._running_plan.monitor_mode = True
        self._plan_queue = QtRePlanQueue(model)
        self._plan_queue.monitor_mode = True
        self._plan_history = QtRePlanHistory(model)
        self._plan_history.monitor_mode = True
        self._console_monitor = QtReConsoleMonitor(model)

        vbox = QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addWidget(self._running_plan)
        self._frame_1.setLayout(vbox)

        vbox = QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addWidget(self._plan_queue)
        self._frame_2.setLayout(vbox)

        vbox = QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addWidget(self._plan_history)
        self._frame_3.setLayout(vbox)

        vbox = QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addWidget(self._console_monitor)
        self._frame_4.setLayout(vbox)

        h = self.sizeHint().height()
        self.setSizes([h, 2 * h, 2 * h, h])


class QtRunEngineManager_Monitor(QWidget):
    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(QtReManagerConnection(model))
        hbox.addWidget(QtReStatusMonitor(model))
        hbox.addStretch()
        vbox.addLayout(hbox)

        vbox.addWidget(QtOrganizeQueueWidgets(model), stretch=2)

        self.setLayout(vbox)


class QtRunEngineManager_Editor(QWidget):
    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(QtReEnvironmentControls(model))
        hbox.addWidget(QtReQueueControls(model))
        hbox.addWidget(QtReExecutionControls(model))
        hbox.addWidget(QtReStatusMonitor(model))

        hbox.addStretch()
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        vbox1 = QVBoxLayout()

        # Register plan editor (opening plans in the editor by double-clicking the plan in the table)
        pe = QtRePlanEditor(model)
        pq = QtRePlanQueue(model)
        pq.registered_item_editors.append(pe.edit_queue_item)

        vbox1.addWidget(pe, stretch=1)
        vbox1.addWidget(pq, stretch=1)
        hbox.addLayout(vbox1)
        vbox2 = QVBoxLayout()
        vbox2.addWidget(QtReRunningPlan(model), stretch=1)
        vbox2.addWidget(QtRePlanHistory(model), stretch=2)
        hbox.addLayout(vbox2)
        vbox.addLayout(hbox)
        self.setLayout(vbox)


class QtViewer(QTabWidget):
    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model

        self.setTabPosition(QTabWidget.West)

        self._re_manager_monitor = QtRunEngineManager_Monitor(model.run_engine)
        self.addTab(self._re_manager_monitor, "Monitor Queue")

        self._re_manager_editor = QtRunEngineManager_Editor(model.run_engine)
        self.addTab(self._re_manager_editor, "Edit and Control Queue")



import gc

from qtpy.QtWidgets import (
    QSizePolicy,
    QTabWidget,
    QWidget,
    QVBoxLayout,
)
from qtpy.QtCore import Signal, QObject
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)
import matplotlib.figure

from ..models.plot_specs import Figure, FigureList
from .._matplotlib_axes import MatplotlibAxes
from ..utils.event import Event
from ..utils.dict_view import DictView


def _initialize_matplotlib():
    "Set backend to Qt5Agg and import pyplot."
    import matplotlib

    matplotlib.use("Qt5Agg")  # must set before importing matplotlib.pyplot
    import matplotlib.pyplot  # noqa


class ThreadsafeMatplotlibAxes(QObject, MatplotlibAxes):
    """
    This overrides the a connect method in MatplotlibAxes to bounce callbacks
    through Qt Signals and Slots so that callbacks run form background threads
    do not run amok.
    """

    __callback_event = Signal(object, Event)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def handle_callback(callback, event):
            callback(event)

        self.__callback_event.connect(handle_callback)

    def connect(self, emitter, callback):
        emitter.connect(lambda event: self.__callback_event.emit(callback, event))


class QtFigures(QTabWidget):
    """
    A Jupyter (ipywidgets) view for a FigureList model.
    """

    __callback_event = Signal(object, Event)

    def __init__(self, model: FigureList, parent=None):
        _initialize_matplotlib()
        super().__init__(parent)
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self._on_close_tab_requested)
        self.resize(self.sizeHint())

        self.model = model
        # Map Figure UUID to widget with QtFigureTab
        self._figures = {}

        for figure_spec in model:
            self._add_figure(figure_spec)
        self._threadsafe_connect(model.events.added, self._on_figure_added)
        self._threadsafe_connect(model.events.removed, self._on_figure_removed)
        self._threadsafe_connect(model.events.active_index, self._on_active_index_changed)

        # This setup for self._threadsafe_connect.

        def handle_callback(callback, event):
            callback(event)

        self.__callback_event.connect(handle_callback)

        self.currentChanged.connect(self._on_tab_changed)

    def sizeHint(self):
        size_hint = super().sizeHint()
        size_hint.setWidth(700)
        size_hint.setHeight(500)
        return size_hint

    @property
    def figures(self):
        "Read-only access to the mapping Figure UUID -> QtFigure"
        return DictView(self._figures)

    def _threadsafe_connect(self, emitter, callback):
        """
        A threadsafe method for connecting to models.

        For example, instead of

        >>> model.events.addded.connect(callback)

        use

        >>> self._threadsafe_connect(model.events.added, callback)
        """
        emitter.connect(lambda event: self.__callback_event.emit(callback, event))

    def _on_close_tab_requested(self, index):
        # When closing is initiated from the view, remove the associated
        # model.
        widget = self.widget(index)
        self.model.remove(widget.model)

    def _on_figure_added(self, event):
        figure_spec = event.item
        self._add_figure(figure_spec)

    def _add_figure(self, figure_spec):
        "Add a new tab with a matplotlib Figure."
        tab = QtFigure(figure_spec, parent=self)
        self.addTab(tab, figure_spec.short_title or figure_spec.title)
        self._figures[figure_spec.uuid] = tab
        # Update the tab title when short_title changes (or, if short_title is
        # None, when title changes).
        self._threadsafe_connect(figure_spec.events.title, self._on_title_changed)
        self._threadsafe_connect(figure_spec.events.short_title, self._on_short_title_changed)

    def _on_figure_removed(self, event):
        "Remove the associated tab and close its canvas."
        figure_spec = event.item
        widget = self._figures[figure_spec.uuid]
        index = self.indexOf(widget)
        self.removeTab(index)
        widget.close_figure()
        del widget
        gc.collect()
        del self._figures[figure_spec.uuid]

    def _on_short_title_changed(self, event):
        "This sets the tab title."
        figure_spec = event.figure_spec
        widget = self._figures[figure_spec.uuid]
        index = self.indexOf(widget)
        # Fall back to title if short_title is being unset.
        if event.value is None:
            self.setTabText(index, figure_spec.title)
        else:
            self.setTabText(index, event.value)

    def _on_title_changed(self, event):
        "This sets the tab title only if short_title is None."
        figure_spec = event.figure_spec
        if figure_spec.short_title is None:
            widget = self._figures[figure_spec.uuid]
            index = self.indexOf(widget)
            self.setTabText(index, event.value)

    def _on_tab_changed(self, index):
        "Update the active_index in the FigureList."
        self.model.active_index = index

    def _on_active_index_changed(self, event):
        "Activate tab based on the updated active index in the FigureList."
        self.setCurrentIndex(event.value)


class QtFigure(QWidget):
    """
    A Qt view for a Figure model. This always contains one Figure.
    """

    def __init__(self, model: Figure, parent=None):
        _initialize_matplotlib()
        super().__init__(parent)
        self.model = model
        self.figure = matplotlib.figure.Figure()
        self.figure.set_tight_layout(True)
        # TODO Let Figure give different options to subplots here,
        # but verify that number of axes created matches the number of axes
        # specified.
        self.axes_list = list(self.figure.subplots(len(model.axes), squeeze=False).ravel())

        self.figure.suptitle(model.title)
        self._axes = {}
        for axes_spec, axes in zip(model.axes, self.axes_list):
            self._axes[axes_spec.uuid] = ThreadsafeMatplotlibAxes(model=axes_spec, axes=axes)
        canvas = FigureCanvas(self.figure)
        canvas.setMinimumWidth(640)
        canvas.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        canvas.updateGeometry()
        canvas.setParent(self)
        toolbar = NavigationToolbar(canvas, parent=self)

        layout = QVBoxLayout()
        layout.addWidget(canvas)
        layout.addWidget(toolbar)
        self.setLayout(layout)
        self.resize(self.sizeHint())

        model.events.title.connect(self._on_title_changed)
        # The Figure model does not currently allow axes to be added or
        # removed, so we do not need to handle changes in model.axes.

    def sizeHint(self):
        size_hint = super().sizeHint()
        size_hint.setWidth(700)
        size_hint.setHeight(500)
        return size_hint

    @property
    def axes(self):
        "Read-only access to the mapping Axes UUID -> MatplotlibAxes"
        return DictView(self._axes)

    def _on_title_changed(self, event):
        self.figure.suptitle(event.value)
        self._redraw()

    def _redraw(self):
        "Redraw the canvas."
        # Schedule matplotlib to redraw the canvas at the next opportunity, in
        # a threadsafe fashion.
        self.figure.canvas.draw_idle()

    def close_figure(self):
        self.figure.canvas.close()



from datetime import datetime, timedelta
import time
import pytest
from qtpy.QtCore import QDateTime
from ...qt.search import QtSearchInput, ADA_LOVELACE_BIRTHDAY, as_qdatetime
from ...models.search import SearchInput, LOCAL_TIMEZONE


def as_datetime(qdatetime):
    assert isinstance(qdatetime, QDateTime)
    return QDateTime.toPython(qdatetime).replace(tzinfo=LOCAL_TIMEZONE)


def datetime_round_trip(dt):
    return as_datetime(as_qdatetime(dt))


def test_datetime_and_QDateTime_round_trip():
    """
    Converting datetime -> QDateTime -> datetime should round trip.

    Test on a recent time (now) and an old time (ADA_LOVELACE_BIRTHDAY).
    """
    # Current time with only seconds resolution
    now = datetime.now(LOCAL_TIMEZONE).replace(microsecond=0)
    assert datetime_round_trip(now) == now
    assert datetime_round_trip(ADA_LOVELACE_BIRTHDAY) == ADA_LOVELACE_BIRTHDAY


def test_initial_state(qtbot):
    "Check that 'All' is checked and since/until are set as expected."
    model = SearchInput()
    view = QtSearchInput(model)
    TOLERANCE = timedelta(seconds=10)

    # Check state of model...
    assert model.since == ADA_LOVELACE_BIRTHDAY
    # ...and view.
    actual_since = as_datetime(view.since_widget.dateTime())
    assert actual_since == ADA_LOVELACE_BIRTHDAY

    assert model.until == timedelta(0)
    actual_until = as_datetime(view.until_widget.dateTime())
    assert actual_until - datetime.now(LOCAL_TIMEZONE) < TOLERANCE

    # Verify that 'All' radio button is checked.
    assert view.all_widget.isChecked()


def test_initial_state_with_fields(qtbot):
    "Check that QLineEdits show for each field."
    model = SearchInput(fields=["field_1", "field_2"])
    view = QtSearchInput(model)

    assert view.field_text_edit
    assert list(view.field_text_edit.keys()) == model.fields
    for field, text_edit in view.field_text_edit.items():
        assert text_edit.text() == ""


def test_time_input_absolute_from_model(qtbot):
    "Set the model to a datetime and verify that the view updates."
    model = SearchInput()
    view = QtSearchInput(model)

    # Set since.
    expected_since = datetime(1980, 2, 2, tzinfo=LOCAL_TIMEZONE)
    model.since = expected_since
    assert model.since == expected_since
    # Set with naive datetime, should be silently converted to local time.
    model.since = expected_since.replace(tzinfo=None)  # naive
    assert model.since == expected_since
    actual_since = as_datetime(view.since_widget.dateTime())
    assert actual_since == expected_since

    # Set until.
    expected_until = datetime(1986, 11, 15, tzinfo=LOCAL_TIMEZONE)
    model.until = expected_until
    assert model.until == expected_until
    # Set with naive datetime, should be silently converted to local time.
    model.until = expected_until.replace(tzinfo=None)  # naive
    assert model.until == expected_until
    actual_until = as_datetime(view.until_widget.dateTime())
    assert actual_until == expected_until


def test_time_input_absolute_from_datetime_picker(qtbot):
    "Set the view QDateTime inputs and verify that the model updates."
    model = SearchInput()
    view = QtSearchInput(model)

    # Set since.
    expected_since = datetime(1980, 2, 2, tzinfo=LOCAL_TIMEZONE)
    view.since_widget.setDateTime(as_qdatetime(expected_since))
    actual_since = as_datetime(view.since_widget.dateTime())
    assert actual_since == expected_since

    # Set until.
    expected_until = datetime(1986, 11, 15, tzinfo=LOCAL_TIMEZONE)
    view.until_widget.setDateTime(as_qdatetime(expected_until))
    actual_until = as_datetime(view.until_widget.dateTime())
    assert actual_until == expected_until


# timedeltas and attribute name of corresponding QRadioButton on QtSearchInput
DELTAS_AND_BUTTONS = [
    (timedelta(hours=-1), "hour_widget"),
    (timedelta(days=-1), "today_widget"),
    (timedelta(weeks=-1), "week_widget"),
    (timedelta(days=-30), "month_widget"),
    (timedelta(days=-365), "year_widget"),
]


@pytest.mark.parametrize("delta, radio_button", DELTAS_AND_BUTTONS)
def test_time_input_relative_from_model(qtbot, delta, radio_button):
    "Set the model to a timedelta and verify that the view updates."
    model = SearchInput()
    view = QtSearchInput(model)
    TOLERANCE = timedelta(seconds=1)

    # Set since.
    expected_since = datetime.now(LOCAL_TIMEZONE) + delta
    model.since = delta
    actual_since = as_datetime(view.since_widget.dateTime())
    assert abs(actual_since - expected_since) < TOLERANCE

    # And until updates automatically.
    expected_until = datetime.now(LOCAL_TIMEZONE)
    actual_until = as_datetime(view.until_widget.dateTime())
    assert abs(actual_until - expected_until) < TOLERANCE

    # And a radio button is checked.
    radio_button_widget = getattr(view, radio_button)
    assert radio_button_widget.isChecked()


@pytest.mark.parametrize("delta, radio_button", DELTAS_AND_BUTTONS)
def test_time_input_from_radio_buttons(qtbot, delta, radio_button):
    "Check the radio buttons in the view and verify that the model updates."
    model = SearchInput()
    view = QtSearchInput(model)
    TOLERANCE = timedelta(seconds=1)

    # Set radio button.
    radio_button_widget = getattr(view, radio_button)
    radio_button_widget.setChecked(True)

    # Check since.
    expected_since = datetime.now(LOCAL_TIMEZONE) + delta
    actual_since = as_datetime(view.since_widget.dateTime())
    assert abs(actual_since - expected_since) < TOLERANCE

    # Check until.
    expected_until = datetime.now(LOCAL_TIMEZONE)
    actual_until = as_datetime(view.until_widget.dateTime())
    assert abs(actual_until - expected_until) < TOLERANCE

    # Additionally, when the model reloads new results, the QDateTime widets
    # should update, reevaluating their relative times with respect to the
    # current time.
    time.sleep(1)
    model.request_reload()

    # Since and until should update with respect to current time.
    new_actual_since = as_datetime(view.since_widget.dateTime())
    new_actual_until = as_datetime(view.until_widget.dateTime())
    assert actual_since != new_actual_since
    assert actual_until != new_actual_until

    # And button should still be checked.
    assert radio_button_widget.isChecked()



from ipywidgets import widgets
import ipympl.backend_nbagg
import matplotlib.figure

from ..models.plot_specs import Figure, FigureList
from .._matplotlib_axes import MatplotlibAxes as _MatplotilbAxes
from ..utils.dict_view import DictView


def _initialize_mpl():
    "Set backend to ipympl and import pyplot."
    import matplotlib

    matplotlib.use("module://ipympl.backend_nbagg")  # must set before importing matplotlib.pyplot
    # must import matplotlib.pyplot here because bluesky.utils.during_task
    # expects it to be imported
    import matplotlib.pyplot as plt  # noqa


class JupyterAxes(_MatplotilbAxes):
    # We need to turn `draw_idle` into "draw now" because the way
    # that `draw_idle` is implemented requires a round-trip
    # communication with the js front end (from the Python side we say
    # "dear JS, when you want you to ask us for an update".  The JS then
    # sends back a message asking for the the figure to be rendered.
    # This means we will not overload the frontend with more updates
    # than it wants.
    #
    # However, when a notebook cell is executing, the zmq loop that
    # processes messages from the front end is blocked so we never see
    # the request for a re-draw. The figure looks "dead".
    #
    # We do not see this problem with Qt because the default "during
    # task" of the RunEngine spins the Qt event loop while we wait on the
    # `_run` task in a background thread so the `draw_idle`
    # implementation throws a signal on the Qt event loop which gets
    # promptly serviced.  In contrast, the default during task when
    # the RE is in the notebook is event.wait which simply blocks
    # until the `_run` task finishes, hence the "broken" behavior.
    def draw_idle(self):
        self.axes.figure.canvas.draw()


class JupyterFigures(widgets.Tab):
    """
    A Jupyter (ipywidgets) view for a FigureList model.
    """

    def __init__(self, model: FigureList, *args, **kwargs):
        _initialize_mpl()
        super().__init__(*args, **kwargs)
        self.model = model
        # Map Figure UUID to widget with JupyterFigureTab
        self._figures = {}

        for figure_spec in model:
            self._add_figure(figure_spec)
        self.model.events.added.connect(self._on_figure_added)
        self.model.events.removed.connect(self._on_figure_removed)

    @property
    def figures(self):
        "Read-only access to the mapping Figure UUID -> JupyterFigure"
        return DictView(self._figures)

    def _on_figure_added(self, event):
        figure_spec = event.item
        self._add_figure(figure_spec)

    def _add_figure(self, figure_spec):
        "Add a new tab with a matplotlib Figure."
        tab = _JupyterFigureTab(figure_spec, parent=self)
        self._figures[figure_spec.uuid] = tab
        self.children = (*self.children, tab)
        index = len(self.children) - 1
        self.set_title(index, figure_spec.title)
        figure_spec.events.title.connect(self._on_title_changed)
        figure_spec.events.short_title.connect(self._on_short_title_changed)
        # Workaround: If the tabs are cleared and then children are added
        # again, no tab is selected.
        if index == 0:
            self.selected_index = 0

    def _on_figure_removed(self, event):
        "Remove the associated tab and close its canvas."
        figure_spec = event.item
        widget = self._figures[figure_spec.uuid]
        children = list(self.children)
        children.remove(widget)
        self.children = tuple(children)
        widget.close_figure()
        del self._figures[figure_spec.uuid]

    def _on_short_title_changed(self, event):
        "This sets the tab title."
        figure_spec = event.figure_spec
        widget = self._figures[figure_spec.uuid]
        index = self.children.index(widget)
        # Fall back to title if short_title is being unset.
        if event.value is None:
            self.set_title(index, figure_spec.title)
        else:
            self.set_title(index, event.value)

    def _on_title_changed(self, event):
        "This sets the tab title only if short_title is None."
        figure_spec = event.figure_spec
        if figure_spec.short_title is None:
            widget = self._figures[figure_spec.uuid]
            index = self.children.index(widget)
            self.set_title(index, event.value)

    def on_close_tab_requested(self, model):
        # When closing is initiated from the view, remove the associated
        # model.
        self.model.remove(model)


class JupyterFigure(widgets.HBox):
    """
    A Jupyter view for a Figure model. This always contains one Figure.
    """

    def __init__(self, model: Figure):
        _initialize_mpl()
        super().__init__()
        self.model = model
        self.figure = matplotlib.figure.Figure()
        # TODO Let Figure give different options to subplots here,
        # but verify that number of axes created matches the number of axes
        # specified.
        self.axes_list = list(self.figure.subplots(len(model.axes), squeeze=False).ravel())
        self.figure.suptitle(model.title)
        self._axes = {}
        for axes_spec, axes in zip(model.axes, self.axes_list):
            self._axes[axes_spec.uuid] = JupyterAxes(model=axes_spec, axes=axes)
        # This updates the Figure's internal state, setting its canvas.
        canvas = ipympl.backend_nbagg.Canvas(self.figure)
        label = "Figure"
        # this will stash itself on the canvas
        ipympl.backend_nbagg.FigureManager(canvas, 0)
        self.figure.set_label(label)
        self.children = (self.figure.canvas,)

        model.events.title.connect(self._on_title_changed)

        # By "resizing" (even without actually changing the size) we bump the
        # ipympl machinery that sets up frontend--backend communication and
        # starting displaying data from the figure. Without this, the figure
        # *widget* displays instantly but the actual *plot* (the PNG data sent from
        # matplotlib) is not displayed until cell execution completes.
        _, _, width, height = self.figure.bbox.bounds
        self.figure.canvas.manager.resize(width, height)
        self.figure.canvas.draw_idle()

        # The Figure model does not currently allow axes to be added or
        # removed, so we do not need to handle changes in model.axes.

    @property
    def axes(self):
        "Read-only access to the mapping Axes UUID -> JupyterAxes"
        return DictView(self._axes)

    def _on_title_changed(self, event):
        self.figure.suptitle(event.value)
        self._redraw()

    def _redraw(self):
        "Redraw the canvas."
        self.figure.canvas.draw()

    def close_figure(self):
        self.figure.canvas.close()


class _JupyterFigureTab(widgets.HBox):
    """
    A tab in a widgets.Tab container that contains a JupyterFigure.

    This is aware of its parent in order to support tab-closing.
    """

    def __init__(self, model: Figure, parent):
        super().__init__()
        self.model = model
        self.parent = parent
        self.button = widgets.Button(description="Close")
        self.button.on_click(lambda _: self.parent.on_close_tab_requested(self.model))
        self._jupyter_figure = JupyterFigure(model)
        self.children = (self._jupyter_figure, self.button)

        # Pass-through accessors to match the API of QtFigure, which has/needs
        # one less layer.
        self.figure = self._jupyter_figure.figure

    def close_figure(self):
        # Pass through toe JupyterFigure instance.
        return self._jupyter_figure.close_figure()

    @property
    def axes(self):
        "Read-only access to the mapping Axes UUID -> JupyterAxes"
        return DictView(self._jupyter_figure.axes)



"""
Use within IPython like

ipython --gui=qt

In [1]: %run -m bluesky_widgets.examples.ipy_qt_figures
"""
from bluesky import RunEngine
from bluesky.plans import scan
from ophyd.sim import motor, det

from bluesky_widgets.utils.streaming import stream_documents_into_runs
from bluesky_widgets.models.auto_plot_builders import AutoLines
from bluesky_widgets.qt.figures import QtFigures
from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog

model = AutoLines(max_runs=3)
view = QtFigures(model.figures)
view.show()

RE = RunEngine()
RE.subscribe(stream_documents_into_runs(model.add_run))


catalog = get_catalog()
scans = catalog.search({"plan_name": "scan"})
model.add_run(scans[-1], pinned=True)


def plan():
    for i in range(1, 5):
        yield from scan([det], motor, -1, 1, 1 + 2 * i)


RE(plan())



def generate_thumbnail(run):
    """
    Return one thumbnail image to represent a Run if any image-like data can be found.

    Logs details to the stderr.

    Parameters
    ----------
    run: BlueskyRun

    Returns
    -------
    image: np.array, None
    """

    uid = run.metadata["start"]["uid"]
    if "primary" in run:
        stream_name = "primary"
    elif list(run):
        # Just grab the first stream.
        stream_name = list(run)[0]
    else:
        print(f"No image data found in Run {uid:.8}")
    dataset = run[stream_name].to_dask()

    # Find the first column that looks like an image.
    # Grab a slice from the middle because that is most likely to be interesting.
    for column in dataset:
        xarr = dataset[column]
        if xarr.ndim == 3:  # a column of single images
            image = dataset[column][xarr.shape[0] // 2]
            break
        elif xarr.ndim == 4:  # a column of stacks of images (like Area Detector gives)
            image = dataset[column][xarr.shape[0] // 2, xarr.shape[1] // 2]
            break
    else:
        print(f"No image data found in Run {uid:.8}")
        return None

    print(f"Found {column!r} from Run {uid}")
    return image



"""
Search for runs and visualize their data.

This example can be run alone as

$ python -m bluesky_widgets.examples.qt_viewer_with_search

or with the data streaming utility which will print an address to connect to

$ python -m bluesky_widgets.examples.utils.stream_data_zmq
Connect a consumer to localhost:XXXXX

python -m bluesky_widgets.examples.qt_viewer_with_search localhost:XXXXX
"""
from bluesky_widgets.models.auto_plot_builders import AutoLines
from bluesky_widgets.jupyter.figures import JupyterFigures


class ExampleApp:
    """
    A user-facing model composed with a Qt widget and window.

    A key point here is that the model `searches` is public and can be
    manipuated from a console, but the view `_window` and all Qt-related
    components are private. The public `show()` and `close()` methods are the
    only view-specific actions that are exposed to the user. Thus, this could
    be implemented in another UI framework with no change to the user-facing
    programmatic interface.
    """

    def __init__(self, *, show=True, title="Example App"):
        super().__init__()
        self.title = title
        self.model = AutoLines(max_runs=3)
        self._widget = JupyterFigures(self.model.figures)

    def _ipython_display_(self, *args, **kwargs):
        "When this object is displayed by Jupyter, display its widget."
        return self._widget._ipython_display_(*args, **kwargs)


def listen_for_data(app, address):
    # Optional: Receive live streaming data.
    from bluesky_widgets.jupyter.zmq_dispatcher import RemoteDispatcher
    from bluesky_widgets.utils.streaming import stream_documents_into_runs

    dispatcher = RemoteDispatcher(address)
    dispatcher.subscribe(stream_documents_into_runs(app.model.add_run))
    dispatcher.start()  # launches process and thread
    return dispatcher.stop



"""
Search for runs and visualize their data.

This example can be run alone as

$ python -m bluesky_widgets.examples.advanced.qt_viewer_with_search

or with the data streaming utility which will print an address to connect to

$ python -m bluesky_widgets.examples.utils.stream_data_zmq
Connect a consumer to localhost:XXXXX

python -m bluesky_widgets.examples.advanced.qt_viewer_with_search localhost:XXXXX
"""
from bluesky_widgets.qt import Window
from bluesky_widgets.qt import gui_qt
from bluesky_widgets.models.search import SearchList, Search
from bluesky_widgets.models.auto_plot_builders import AutoLines
from bluesky_widgets.qt.search import QtSearches
from bluesky_widgets.qt.figures import QtFigures
from bluesky_widgets.utils.event import Event
from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog
from bluesky_widgets.examples.utils.add_search_mixin import columns
from qtpy.QtWidgets import QWidget, QPushButton, QHBoxLayout, QVBoxLayout


class SearchListWithButton(SearchList):
    """
    A SearchList model with a method to handle a click event.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.events.add(view=Event)


class QtSearchListWithButton(QWidget):
    """
    A view for SearchListWithButton.

    Combines the QtSearches widget with a button that processes the selected Runs.
    """

    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(QtSearches(model))

        go_button = QPushButton("View Selected Runs")
        layout.addWidget(go_button)
        go_button.clicked.connect(self.model.events.view)


class SearchAndView:
    def __init__(self, searches, viewer):
        self.searches = searches
        self.viewer = viewer
        self.searches.events.view.connect(self._on_view)

    def _on_view(self, event):
        for uid, run in self.searches.active.selection_as_catalog.items():
            self.viewer.add_run(run, pinned=True)


class QtSearchAndView(QWidget):
    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        layout = QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(QtSearchListWithButton(model.searches))
        layout.addWidget(QtFigures(model.viewer.figures))


class ExampleApp:
    """
    A user-facing model composed with a Qt widget and window.

    A key point here is that the model `searches` is public and can be
    manipuated from a console, but the view `_window` and all Qt-related
    components are private. The public `show()` and `close()` methods are the
    only view-specific actions that are exposed to the user. Thus, this could
    be implemented in another UI framework with no change to the user-facing
    programmatic interface.
    """

    def __init__(self, *, show=True, title="Example App"):
        super().__init__()
        self.title = title
        self.searches = SearchListWithButton()
        self.viewer = AutoLines(max_runs=3)
        self.model = SearchAndView(self.searches, self.viewer)
        widget = QtSearchAndView(self.model)
        self._window = Window(widget, show=show)

        # Initialize with a two search tabs: one with some generated example data...
        self.searches.append(Search(get_catalog(), columns=columns))
        # ...and one listing any and all catalogs discovered on the system.
        from databroker import catalog

        self.model.searches.append(Search(catalog, columns=columns))

    def show(self):
        """Resize, show, and raise the window."""
        self._window.show()

    def close(self):
        """Close the window."""
        self._window.close()


def main(argv):
    print(__doc__)

    with gui_qt("Example App"):
        app = ExampleApp()

        # Optional: Receive live streaming data.
        if len(argv) > 1:
            from bluesky_widgets.qt.zmq_dispatcher import RemoteDispatcher
            from bluesky_widgets.utils.streaming import (
                stream_documents_into_runs,
            )

            address = argv[1]
            dispatcher = RemoteDispatcher(address)
            dispatcher.subscribe(stream_documents_into_runs(app.viewer.add_run))
            dispatcher.start()

        # We can access and modify the model as in...
        len(app.searches)
        app.searches[0]
        app.searches.active  # i.e. current tab
        app.searches.active.input.since  # time range
        app.searches.active.input.until
        app.searches.active.results
        app.searches.active.selection_as_catalog
        app.searches.active.selected_uids


if __name__ == "__main__":
    import sys

    main(sys.argv)
